#include "para.h"

/*
 * fds must be a pointer to three integers, corresponding to stdin,
 * stdout and stderr respectively.
 * fd[i] < 0 : leave fd alone
 * fd[i] == 0: dup fd to /dev/null
 * fd[i] > 0 : dup fd and return fileno in fd[i]
 */
int para_exec(pid_t *pid, const char *file, char *const args[], int *fds)
{
	int in[2] = {-1, -1}, out[2] = {-1, -1}, err[2] = {-1, -1},
		null = -1; /* ;) */

	if (fds[0] > 0 && pipe(in) < 0)
		goto err_out;
	if (fds[1] > 0 && pipe(out) < 0)
		goto err_out;
	if (fds[2] > 0 && pipe(err) < 0)
		goto err_out;
	if (!fds[0] || !fds[1] || !fds[2]) {
		null = open("/dev/null", O_RDONLY);
		if (null < 0)
			goto err_out;
	}
	if ((*pid = fork()) < 0)
		exit(EXIT_FAILURE);
	if (!(*pid)) { /* child */
		int i;
		/* close unneeded fds */
		for (i = 3; i < 256; i++)
			if (i != in[0]
				&& i != in[1]
				&& i != out[0]
				&& i != out[1]
				&& i != err[0]
				&& i != err[1]
				&& i != null
			)
				close(i);
		if (fds[0] >= 0) {
			if (fds[0]) {
				close(in[1]);
				if (in[0] != STDIN_FILENO)
					dup2(in[0], STDIN_FILENO);
			} else
				dup2(null, STDIN_FILENO);
		}
		if (fds[1] >= 0) {
			if (fds[1]) {
				close(out[0]);
				if (out[1] != STDOUT_FILENO)
					dup2(out[1], STDOUT_FILENO);
			} else
				dup2(null, STDOUT_FILENO);
		}
		if (fds[2] >= 0) {
			if (fds[2]) {
				close(err[0]);
				if (err[1] != STDERR_FILENO)
					dup2(err[1], STDERR_FILENO);
			} else
				dup2(null, STDERR_FILENO);
		}
		if (null >= 0)
			close(null);
		execvp(file, args);
		_exit(EXIT_FAILURE);
	}
	if (fds[0] > 0) {
		close(in[0]);
		*fds = in[1];
	}
	if (fds[1] > 0) {
		close(out[1]);
		*(fds + 1) = out[0];
	}
	if (fds[2] > 0) {
		close(err[1]);
		*(fds + 2) = err[0];
	}
	if (null >= 0)
		close(null);
	return 1;
err_out:
	if (err[0] >= 0)
		close(err[0]);
	if (err[1] >= 0)
		close(err[1]);
	if (out[0] >= 0)
		close(out[0]);
	if (out[1] >= 0)
		close(out[1]);
	if (in[0] >= 0)
		close(in[0]);
	if (in[1] >= 0)
		close(in[1]);
	if (null >= 0)
		close(null);
	return -1;
}

int split_args(char *args, char ***argv_ptr, int delim)
{
	char *p = args;
	char **argv;
	ssize_t n = 0, i;

	while ((p = strchr(p, delim))) {
		p++;
		n++;
	}
	*argv_ptr = malloc((n + 3) * sizeof(char *));
	argv = *argv_ptr;
	i = 0;
	p = args;
//	printf("split_args: a:%s\n", p);
	while (p) {
		argv[i] = p;
		p = strchr(p, delim);
		if (p) {
//		printf("a:%s\n", p);
			*p = '\0';
//		printf("b:\n");
			p++;
			i++;
		}
	}
	argv[n + 1] = NULL;
	return n;
}

int para_exec_cmdline_pid(pid_t *pid, char *cmdline, int *fds)
{
	int argc, ret = -1;
	char **argv, *tmp = strdup(cmdline);

	if (!tmp)
		exit(EXIT_FAILURE);
	argc = split_args(tmp, &argv, ' ');
	if (argv)
		ret = para_exec(pid, argv[0], argv, fds);
	free(tmp);
	return ret;
}

int file_exists(const char *fn)
{
        struct stat statbuf;

        return !stat(fn, &statbuf);
}

/*
 * Check if file is fifo, create if neccessary and open with given
 * flags.
 */
int para_open_fifo(const char *fifoname, const int flags)
{
	int ret, fifo;
	struct stat stat_buf;

	ret = stat(fifoname, &stat_buf);
	if (ret < 0) {
		ret = mknod(fifoname, S_IFIFO | S_IRUSR | S_IWUSR, 42);
		if (ret >= 0)
			ret = stat(fifoname, &stat_buf);
		if (ret < 0)
			return ret;
	}
	if (!S_ISFIFO(stat_buf.st_mode))
		return -1;
	fifo = open(fifoname, flags);
	if (fifo < 0)
		return -1;
	return fifo;
}
